function [feature_order,MIh]=LSFS(x,y,y_type,feature_group,sigma_list,lambda_list,b,fold)
%
% Least-Squares Feature Selection
%
% Order input features according to their dependency on outputs
%
% Usage:
%       [feature_order,MIh]=LSFS(x,y,y_type,feature_group,sigma_list,lambda_list,b)
%
% Input:
%    x             : dx by n input sample matrix
%    y             : dy by n output sample matrix
%    y_type     : if y_type=1, delta kernel is used for y; 
%                 otherwise (or empty) Gaussian kernel is used.
%    feature_group : dx-dimensional vector indicating group indices;
%                    if dx features are grouped, their group indices are specified.
%                    Default is [1:dx].
%    sigma_list    : (candidates of) Gaussian width
%                    If sigma_list is a vector, one of them is selected by cross validation.
%                    If sigma_list is a scalar, this value is used without cross validation
%                    If sigma_list is empty/undefined, Gaussian width is chosen from
%                    some default canditate list by cross validation
%    lambda_list   : Regularization parameter
%                    If lambda_list is a vector, one of them is selected by cross validation.
%                    If lambda_list is a scalar, this value is used without cross validation
%                    If lambda_list is empty, Gaussian width is chosen from
%                    some default canditate list by cross validation
%    b             : Number of Gaussian centers (if empty, b=200 is used);
%
% Output:
%    feature_order : Feature orders 
%                    feature_order(1)  : most dependent
%                    feature_order(end): least dependent
%    MIh           : Estimated squared-loss mutual information for each feature
%
% (c) Taiji Suzuki, Department of Mathematical Informatics, The University of Tokyo, Japan. 
%     Masashi Sugiyama, Department of Compter Science, Tokyo Institute of Technology, Japan.
%     s-taiji@stat.t.u-tokyo.ac.jp
%     sugi@cs.titech.ac.jp,

[dx n] =size(x);
[dy ny] =size(y);
if n~=ny
    error('number of samples of x and y must be the same!!!')
end

if nargin<3 || isempty(y_type)
  y_type=0;
end

if nargin < 4 || isempty(feature_group)
    feature_group = [1:dx]; 
end
if length(feature_group) ~= dx
    error('the length of feature_group must be equal to the dimension of x!!!')    
end

if nargin < 5 || isempty(sigma_list)
  sigma_list=logspace(-2,2,9);
end

if nargin < 6 || isempty(lambda_list)
  lambda_list=logspace(-3,1,9);
end

if nargin < 7 || isempty(b)
  b = 200;
end  

if nargin < 8 || isempty(fold)
  fold = 5;
end  

b=min(n,b);
       
group_list=unique(feature_group);

for group_index=1:length(group_list)
  %Extract a grouped feature
  z=x(find(feature_group==group_list(group_index)),:);

  %Gaussian centers are randomly chosen from samples
  rand_index=randperm(n);
  u=z(:,rand_index(1:b));
  v=y(:,rand_index(1:b));
  
  Phiz_tmp=GaussBasis_sub(z,u)';
  if y_type~=1
    Phiy_tmp=GaussBasis_sub(y,v)';
  end
  
  
  if length(sigma_list)==1 && length(lambda_list)==1
    sigma_chosen=sigma_list;
    lambda_chosen=lambda_list;
    score_cv=-inf;
  else %length(sigma_list)==1 && length(lambda_list)==1
    %%%%%%%%%%%%%%%% Searching Gaussian kernel width `sigma_chosen'
    %%%%%%%%%%%%%%%% and regularization parameter `lambda_chosen' 
    fold_index=[1:fold];
    cv_index=randperm(n);
    cv_split=floor([0:n-1]*fold./n)+1;
    scores_cv=zeros(length(sigma_list),length(lambda_list));
    
    for sigma_index=1:length(sigma_list)
      sigma=sigma_list(sigma_index);
      Phiz_sigma=GaussBasis(Phiz_tmp,sigma);
      if y_type
        Phiy_sigma=DeltaBasis(y,v);
      else
        Phiy_sigma=GaussBasis(Phiy_tmp,sigma);
      end
      Phi_sigma=Phiz_sigma.*Phiy_sigma;
      
      for i=fold_index
        cv_index_tmp=cv_index(cv_split==i);
        HH_cv_Phiz(:,:,i)=Phiz_sigma(:,cv_index_tmp)*Phiz_sigma(:,cv_index_tmp)';
        HH_cv_Phiy(:,:,i)=Phiy_sigma(:,cv_index_tmp)*Phiy_sigma(:,cv_index_tmp)';
        hh_cv(:,:,i)=mean(Phi_sigma(:,cv_index_tmp),2);
        n_cv(i)=length(cv_index_tmp);
      end
      
      for i=fold_index
        cv_index_tr=fold_index(fold_index~=i);
        Hh_cv_tr=sum(HH_cv_Phiz(:,:,cv_index_tr),3) ...
                 .*sum(HH_cv_Phiy(:,:,cv_index_tr),3)/(sum(n_cv(cv_index_tr))^2);
        Hh_cv_te=HH_cv_Phiz(:,:,i).*HH_cv_Phiy(:,:,i)/(n_cv(i)^2);
        hh_cv_tr=mean(hh_cv(:,:,cv_index_tr),3);
        hh_cv_te=hh_cv(:,:,i);
        for lambda_index=1:length(lambda_list)
          lambda=lambda_list(lambda_index);
          alphah_cv=mylinsolve(Hh_cv_tr+lambda*eye(b),hh_cv_tr);
          wh_cv=alphah_cv'*Hh_cv_te*alphah_cv/2-hh_cv_te'*alphah_cv;
          scores_cv(sigma_index,lambda_index)=scores_cv(sigma_index,lambda_index)+wh_cv/fold;
        end % fold
      end % lambda
    end % sigma
    [scores_cv_tmp,lambda_chosen_index]=min(scores_cv,[],2);
    [score_cv,sigma_chosen_index]=min(scores_cv_tmp);
    lambda_chosen=lambda_list(lambda_chosen_index(sigma_chosen_index));
    sigma_chosen=sigma_list(sigma_chosen_index);
  end %length(sigma_list)==1 && length(lambda_list)==1
  
  %%%%%%%%%%%%%%%% Computing the MI estimator `MIh'
  Phiz=GaussBasis(Phiz_tmp,sigma_chosen);
  if y_type
    Phiy=DeltaBasis(y,v);
  else
    Phiy=GaussBasis(Phiy_tmp,sigma_chosen);
  end
  Phi=Phiz.*Phiy;
  Hh=(Phiz*Phiz').*(Phiy*Phiy')/(n^2);
  hh=mean(Phi,2);
  alphah=mylinsolve(Hh+lambda_chosen*eye(b),hh);
%  MIh(group_index)=hh'*alphah-alphah'*Hh*alphah/2-1/2;
%  MIh(group_index)=alphah'*Hh*alphah/2-(mean(Phiz,2).*mean(Phiy,2))'*alphah+1/2;
  MIh(group_index)=hh'*alphah/2-1/2;
end %group_index

[dummy Ind]=sort(MIh,'descend');
feature_order=Ind;
